import contextlib

import matplotlib as mpl
import matplotlib.pyplot as plt

from sukta.utils.tree import apply_vfunc

from .colors import TurboColormap

overriden_styles = ["default"]


@contextlib.contextmanager
def context(style, after_reset: bool = False):
    with mpl.rc_context():
        if after_reset:
            mpl.rcdefaults()
        use(style)
        yield


def use(style):
    TurboColormap.register_turbocmap()

    def _replace_overriden(style):
        if style in overriden_styles:
            return "sukta.plot.styles." + style
        return style

    plt.style.use(apply_vfunc(_replace_overriden, style))



"""Telemetry"""

from pathlib import Path
from typing import Dict, Optional, Union

import numpy as np
from aim import Distribution, Run

KVDict = Dict[str, Union[str, int, float, bool]]
Array = np.ndarray


class Telemetry:
    """Telemetry class for logging metrics and hyperparameters using Aim"""

    def __init__(self, repo: Path, experiment: str):
        self.run = Run(
            repo=str(repo.expanduser().absolute()),
            experiment=experiment,
            log_system_params=True,
            capture_terminal_logs=False,
        )

    def hparams(self, params: KVDict):
        """Log hyperparameters"""
        self.run["hparams"] = params

    def log(
        self,
        metrics: KVDict,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        context: Optional[KVDict] = None,
    ):
        """Log metrics"""
        self.run.track(
            metrics,
            step=step,
            epoch=epoch,
            context=context,
        )

    def distribution(
        self,
        name: str,
        values: Array,
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        context: Optional[KVDict] = None,
        bin_count: int = 64,
    ):
        """Log distribution"""
        d = Distribution(
            distribution=values,
            bin_count=bin_count,
        )
        self.run.track(d, name=name, step=step, epoch=epoch, context=context)

    def close(self):
        """Close the telemetry run"""
        self.run.close()



"""Runner utility"""

import cProfile
import inspect
import io
import os
import pdb
import pstats
import re
import signal
import sys
import time
from functools import singledispatch
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

import cloudpickle as pickle
import dacite
from benedict import benedict
from jsonargparse import ArgumentParser, Namespace

from sukta import sem
from sukta.logging import LOG_LEVEL, Logger
from sukta.telemetry import Telemetry

T = TypeVar("T")


def cache(cache_dir, cache_name=None, cache_argnames=None):
    """provides a decorator to cache the results of any function.
    It pickles the return and stores it to reload if cached_path saved."""

    def decorator(func):
        sig = inspect.signature(func)
        param_names = list(sig.parameters.keys())

        def wrapper(*args, **kwargs):
            cdir = Path(cache_dir)
            if cache_name is None:
                cname = func.__name__
            else:
                cname = cache_name
            if cache_argnames is not None:
                bound_args = sig.bind(*args, **kwargs)
                bound_args.apply_defaults()
                all_args = bound_args.arguments
                cdir = cdir / "$".join(
                    f"{k}:{all_args.get(k, '')}" for k in cache_argnames
                )
            if not cname.endswith(".pkl"):
                cname += ".pkl"
            cdir.mkdir(parents=True, exist_ok=True)
            cache_path = cdir / cname
            if cache_path.exists():
                with open(cache_path, "rb") as f:
                    result = pickle.load(f)
                return result
            result = func(*args, **kwargs)
            with open(cache_path, "wb") as f:
                pickle.dump(result, f)
            return result

        return wrapper

    return decorator


# convert jsonargparse Namespace to dict recursively if needed
@singledispatch
def namespace2dict(namespace):
    return namespace


@namespace2dict.register(Namespace)
def _namespace2dict(namespace: Namespace) -> Dict:
    return {k: namespace2dict(v) for k, v in vars(namespace).items()}


class EnvConfig:
    """Utility class to configure common environment variables for different libraries."""

    @classmethod
    def jax(
        cls,
        preallocate: bool = False,
        mem_fraction: float = 0.75,
        alloc_on_demand: bool = False,
        device_id: int | Tuple[int, ...] | None = None,
    ):
        """configure common jax environment variables"""
        os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = f"{mem_fraction:.2f}"
        if alloc_on_demand:
            os.environ["XLA_PYTHON_CLIENT_ALLOCATOR"] = "platform"
        os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = str(preallocate).lower()
        xla_flags = os.environ.get("XLA_FLAGS", "")
        xla_flags += " --xla_gpu_triton_gemm_any=True"
        os.environ["XLA_FLAGS"] = xla_flags
        if device_id is not None:
            if isinstance(device_id, int):
                device_id = (device_id,)
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, device_id))
            if len(device_id) == 0:
                os.environ["JAX_PLATFORMS"] = "cpu"

    @classmethod
    def mujoco(
        cls, gl: Literal["egl", "osmesa", "glfw"] = "egl", gl_device_id: int = 0
    ):
        if "MUJOCO_GL" not in os.environ:
            os.environ["MUJOCO_GL"] = gl
            if gl == "egl":
                # warning: might not be supported on all cuda-driver versions
                os.environ["MUJOCO_EGL_DEVICE_ID"] = str(gl_device_id)


class Runner:
    EnvConfig = EnvConfig

    def __init__(
        self,
        root_dir: Path = Path("./runs"),
        experiment: Optional[str] = None,
        override: bool = False,
        run_params: Any = {},
        auto_sort: bool = True,
        log_name: str = "run",
        log_level: LOG_LEVEL = "INFO",
        exit_fns: Sequence[Callable] | None = None,
        debug: bool = False,
    ):
        exp_params = (experiment,) if experiment is not None else ()
        if isinstance(run_params, list) or isinstance(run_params, tuple):
            exp_params = (*exp_params, *run_params)
        else:
            exp_params = (*exp_params, run_params)
        self.exp_params = exp_params
        self.auto_sort = auto_sort
        self.run_dir = sem.ResultManager.create_default_path(
            root_dir, exp_params, auto_sort=auto_sort
        )
        if not override:
            assert not self.run_dir.exists(), (
                f"Run directory {self.run_dir} already exists"
            )
        self.log_dir = self.run_dir / "logs"
        self.log_dir.mkdir(parents=True, exist_ok=True)
        self.log_name = log_name
        self.log_level = log_level
        self._exit_fns = exit_fns or []

        self.log = Logger.getLogger(self.log_name, self.run_dir, level=self.log_level)

        repo = root_dir / experiment
        self.telemetry = Telemetry(
            repo=repo,
            experiment=experiment,
        )
        self.exit_hook(self.telemetry.close)
        self.debug = debug

    def get_experiment_dir(self, root_dir: Path):
        """
        utility method to allow specifying slower drives to store checkpoints etc., if needed
        """
        return sem.ResultManager.create_default_path(
            root_dir, self.exp_params, auto_sort=self.auto_sort
        )

    def __enter__(self):
        signal.signal(signal.SIGINT, self.handle_stop_signals)
        signal.signal(signal.SIGTERM, self.handle_stop_signals)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        for fn in self._exit_fns:
            fn()
        if exc_type is not None:
            self.log.exception("An exception occurred")
            if self.debug:
                pdb.post_mortem(traceback)

    def handle_stop_signals(self, signum, frame):
        self.log.warning("Received signal %s, stopping...", signal.Signals(signum).name)
        sys.exit(1)

    @classmethod
    def default_parser(
        cls,
        parser: Optional[ArgumentParser] = None,
        run_dir: Path = "./runs",
        experiment: Optional[str] = None,
        log_name: str = "run",
        log_level: LOG_LEVEL = "INFO",
        debug: bool = False,
        override: bool = False,
    ) -> ArgumentParser:
        """Create a default argument parser for Runner, or add to an existing one."""
        parser = parser or ArgumentParser()
        parser.add_argument("--debug", type=bool, default=debug)
        parser.add_argument("--override", type=bool, default=override)
        parser.add_argument("--run_dir", type=Path, default=run_dir)
        parser.add_argument("--experiment", type=Optional[str], default=experiment)
        parser.add_argument("--log_name", type=str, default=log_name)
        parser.add_argument(
            "--desc",
            type=str,
            default=None,
            help="use to provide a memorable description for the experiment",
        )
        parser.add_argument(
            "--log_level",
            type=LOG_LEVEL,
            default=log_level,
        )
        return parser

    @classmethod
    def from_args(
        cls,
        args: Namespace,
        run_params: Any = {},
        exit_fns: Sequence[Callable] | Callable | None = None,
    ) -> "Runner":
        """Create a Runner instance from command line arguments."""
        if isinstance(exit_fns, Callable):
            exit_fns = [exit_fns]
        runner = cls(
            root_dir=args.run_dir,
            experiment=args.experiment,
            override=args.override,
            log_name=args.log_name,
            log_level=args.log_level,
            debug=args.debug,
            run_params=run_params,
            exit_fns=exit_fns,
        )
        return runner

    def save_args(self, parser: ArgumentParser, args: Namespace):
        """Save the command line arguments to a yaml file in the run directory."""
        with open(self.run_dir / "args.yaml", "w", encoding="utf-8") as f:
            f.write(parser.dump(args))
        if args.desc is not None:
            with open(self.run_dir / "desc.txt", "w", encoding="utf-8") as f:
                f.write(args.desc)

    def save_dict_as_yaml(self, data_name: str, data: Dict):
        """Save a dictionary as a yaml file in the run directory."""
        assert "." not in data_name, ".yaml is automatically appended"
        assert data_name != "args", "args.yaml is reserved for storing run arguments"
        with open(self.run_dir / f"{data_name}.yaml", "w", encoding="utf-8") as f:
            f.write(benedict(data).to_yaml())

    def exit_hook(self, fn: Callable):
        """Register a function to be called when the Runner exits."""
        self._exit_fns.append(fn)

    @classmethod
    def load_args(cls, path) -> Dict:
        """Load command line arguments from a yaml file."""
        return benedict.from_yaml(path)

    @classmethod
    def dataclass_from_dict(
        cls,
        data_class: Type[T],
        data: dacite.data.Data,
        config: dacite.Config | None = None,
    ) -> T:
        if isinstance(data, Namespace):
            data = namespace2dict(data)
        return dacite.from_dict(data_class, data, config)

    def set_trace(self):
        """Set a breakpoint using pdb."""
        pdb.set_trace()

    @staticmethod
    def parse_cstr(cstr):
        """Parse configuration-string (e.g. serialized hyperparameters) to dictionary."""
        replaces, restores = {}, {}
        for c in ["_", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
            sp_chr = f"||{chr(ord('A') + ord(c) - ord('0'))}||"
            replaces[f"\\{c}"] = sp_chr
            restores[sp_chr] = c

        def replace(s: str) -> str:
            for k, v in replaces.items():
                s = s.replace(k, v)
            return s

        def restore(s: str) -> str:
            for k, v in restores.items():
                s = s.replace(k, v)
            return s

        cstr = replace(cstr)
        chunks = cstr.split("_")  # simple underscore to split chunks
        d = {}
        for chunk in chunks:
            if "$" in chunk:
                k, v = chunk.split("$", 1)
                k, v = restore(k), restore(v)
                d[k] = v
            else:
                match = re.match(r"([a-zA-Z\|]+)(.*)", chunk)
                if match:
                    k, v = match.groups()
                    k, v = restore(k), restore(v)
                    try:
                        if "." in v or "e" in v.lower():
                            d[k] = float(v)
                        else:
                            d[k] = int(v)
                    except (ValueError, TypeError):
                        # if conversion fails, test if it's a flag parameter
                        if len(v) == 0:
                            d[k] = True  # flag parameter
                        else:
                            d[k] = v  # keep as string
        return d

    @staticmethod
    def serialize_cstr(cdict: Dict) -> str:
        """Serialize a dictionary to a configuration-string."""
        cstr = ""
        for k, v in cdict.items():
            k = str(k).replace("_", r"\_")
            if isinstance(v, bool) and v is True:
                cstr += f"_{k}"
            else:
                if isinstance(v, float):
                    v = "{:e}".format(v)
                    v = re.sub(
                        r"([eE][+-]?)0+", r"\1", v
                    )  # remove leading zeros in exponent
                    v = re.sub(r"\.0+([eE])", r"\1", v)  # remove .0 in float
                v = str(v).replace("_", r"\_")
                cstr += f"_{k}${v}"
        if cstr.startswith("_"):
            cstr = cstr[1:]
        return cstr


class Profile:
    """Context manager that profiles a section of code"""

    def __init__(self, name: str, sort_by="cumulative"):
        self.name = name

        self.profiler = cProfile.Profile()
        self.sort_by = sort_by
        self.start_time = None

    def __enter__(self):
        self.profiler.enable()
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        total_time = time.time() - self.start_time
        self.profiler.disable()
        s = io.StringIO()
        ps = pstats.Stats(self.profiler, stream=s).sort_stats(self.sort_by)
        ps.print_stats()
        print(f"Profile [{self.name}]:\n{s.getvalue()}")
        print(f"Profile [{self.name}] took {total_time:.3f} seconds")


class Timer:
    """Context manager that times a section of code"""

    def __init__(self, name: str = "", logger: Optional[Logger] = None):
        self.name = name
        self.logger = logger
        self.start_time = None

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        total_time = time.time() - self.start_time
        total_time = Logger.num2pstr(total_time, ".3f")
        name = self.name or "root"
        msg = "[{name}] took {total_time}s"
        if self.logger is not None:
            self.logger.info(msg, name=name, total_time=total_time)
        else:
            print(msg.format(name=name, total_time=total_time))



from typing import Optional, Sequence, Union

import jax.random as jr
from jaxtyping import Key

OKey = Optional[Key]


def split_support_none(key: Union[OKey, Sequence[OKey]], num: int = 2):
    if isinstance(key, Sequence):
        key = key[0]
    if key is None:
        return (None,) * num
    return jr.split(key, num)



from typing import Any, Callable, Dict, List, Optional, Type, Union


def iter_path(
    data: Any,
    prefix: Optional[List[str]] = None,
    delim: Optional[str] = None,
    dict_type: Type = dict,
    list_type: Type = list | tuple,
    match_key: Optional[str] = None,
):
    """Generator to iterate over (key_path, value) of a nested dict/list structure (PyTree)."""
    if prefix is None:
        prefix = []
    if isinstance(data, dict_type):
        for k, v in data.items():
            if k == match_key:
                if delim is None:
                    yield prefix, v
                else:
                    yield delim.join(prefix), v
            else:
                yield from iter_path(
                    v, prefix + [k], delim, dict_type, list_type, match_key
                )
    elif isinstance(data, list_type):
        for i, v in enumerate(data):
            yield from iter_path(
                v, prefix + [str(i)], delim, dict_type, list_type, match_key
            )
    else:
        if match_key is None:
            if delim is None:
                yield prefix, data
            else:
                yield delim.join(prefix), data


def apply_vfunc(
    vfunc: Callable[[Any], Any],
    data: Any,
    dict_type: Type = dict,
    list_type: Type = list | tuple,
    preserve_ret_type: bool = True,
):
    """Apply a function on each leaf of a nested dict/list structure (PyTree)."""
    if isinstance(data, dict_type):
        ret_type = type(data) if preserve_ret_type else dict
        return ret_type(
            {k: apply_vfunc(vfunc, v, dict_type, list_type) for k, v in data.items()}
        )
    elif isinstance(data, list_type):
        ret_type = type(data) if preserve_ret_type else list
        return ret_type([apply_vfunc(vfunc, v, dict_type, list_type) for v in data])
    return vfunc(data)


def apply_pfunc(
    pfunc: Callable[[List[str], Any], Any],
    data: Any,
    prefix: Optional[List[str]] = None,
    delim: Optional[str] = None,
    dict_type: Type = dict,
    list_type: Type = list | tuple,
    preserve_ret_type: bool = True,
):
    """Apply a function on each leaf of a nested dict/list structure (PyTree), takes in path and leaf node as input."""
    if prefix is None:
        prefix = []
    if isinstance(data, dict_type):
        ret_type = type(data) if preserve_ret_type else dict
        return ret_type(
            {
                k: apply_pfunc(pfunc, v, prefix + [k], delim, dict_type, list_type)
                for k, v in data.items()
            }
        )
    elif isinstance(data, list_type):
        ret_type = type(data) if preserve_ret_type else dict
        return ret_type(
            [
                apply_pfunc(pfunc, v, prefix + [str(i)], delim, dict_type, list_type)
                for i, v in enumerate(data)
            ]
        )
    if delim is None:
        return pfunc(prefix, data)
    else:
        return pfunc(delim.join(prefix), data)


def type_cond_func(funcs: Dict[Type, Callable], fallback: Optional[Callable] = None):
    """Apply function based on the type of the input data."""

    def f(x):
        for tp, func in funcs.items():
            if isinstance(x, tp):
                return func(x)
        if fallback is not None:
            return fallback(x)
        else:
            raise ValueError(f"no function support for type {type(x)}")

    return f


def resolve_path(path: Union[str, List[str]], delim: str):
    """Convert path to list of strings if it is a string."""
    if isinstance(path, str):
        return path.split(delim)
    return path


def set_path(
    data: Any,
    path: str | List[str],
    value,
    dict_type: Type = dict,
    list_type: Type = list | tuple,
    delim: str = ".",
):
    """Set value at the given path of a nested dict/list structure (PyTree)."""
    x = data
    path = resolve_path(path, delim)
    path_len = len(path)
    for i, p in enumerate(path):
        # check if nonnegative integer
        if p.isdigit():
            p = int(p)
            if not isinstance(x, list_type):
                raise ValueError(f"Expect list/tuple type at {p}, got {type(x)}")
            if p >= len(x):
                raise IndexError(f"Index {p} out of range for list of length {len(x)}")
        else:
            if not isinstance(x, dict_type):
                raise ValueError(f"Expect dict type at {p}, got {type(x)}")
            if p not in x:
                raise KeyError(f"Key {p} not found in dict {x}")
        if i == path_len - 1:
            x[p] = value
            break
        x = x[p]
    return data


def get_path(data: Any, path: str | List[str], delim: str = "."):
    """Get value at the given path of a nested dict/list structure (PyTree)."""
    x = data
    path = resolve_path(path, delim)
    for p in path:
        x = x[p]
    return x


def has_path(data: Any, path: str | List[str], delim: str = "."):
    """Check if the given path exists in a nested dict/list structure (PyTree)."""
    x = data
    path = resolve_path(path, delim)
    for p in path:
        if p.isdigit():
            p = int(p)
            if not isinstance(x, (list, tuple)):
                return False
            if p >= len(x):
                return False
        elif p not in x:
            return False
        x = x[p]
    return True



"""
Common type conversion from string, use to parse arguments.
"""

from collections.abc import Callable
from typing import Any, Optional, Union


def bool_type(x: str) -> bool:
    """Convert a valid string to a boolean.
    Valid strings are: "0", "1", "true" and "false".
    The conversion is case-insensitive.
    """
    x = x.strip()
    if x == "1" or x.lower() == "true":
        return True
    if x == "0" or x.lower() == "false":
        return False

    raise ValueError("The input cannot be converted to a boolean")


def none_or_type(default_type: Callable) -> Callable:
    """Return a new type function.
    The new functions parses the "None" string (in a case insensitive fashion),
    otherwise calls a predefined function default_type.
    """

    def new_type_fn(x: str) -> Optional[Any]:
        x = x.strip()
        if x.lower() == "none":
            return None
        return default_type(x)

    return new_type_fn


def unit_float_or_positive_integer(x: str, valid_zero=True) -> Union[int, float]:
    """Convert a string to a float within the open interval (0, 1) or to a positive
    integer.
    Zero is considered a valid integer input depending on the value of valid_zero.
    This type of input is utilized in many libraries where it either represents a
    fraction (i.e. a value between 0 and 1), or a an absolute count of objects.
    """
    x = float(x)

    if x < 0:
        raise ValueError("Invalid input ", x, " : less or equal than 0.")
    if x == 0 and not valid_zero:
        raise ValueError(
            "Invalid zero input. If you wish to accept zero as input, set valid_zero to True."
        )
    if x > 1 and not x.is_integer():
        raise ValueError("Can only accept float values in the range (0, 1).")

    if x.is_integer():
        return int(x)
    return x



from __future__ import annotations

import collections
import re
from pathlib import Path
from typing import Any, Dict, List, Optional

import pandas as pd

import sukta.utils.tree as tree
from sukta.sem.structure_parsing import recursive_folder_parsing


class ResultManager:
    """A manager for experimental results.
    It takes care of collecting results organized in different folders.
    """

    def __init__(
        self,
        root_dir: Path,
        parsers: List[re.Pattern],
        dataframe: Optional[pd.DataFrame] = None,
    ):
        """Initialize the ResultManager.

        :param root_dir: root directory.
        :param parsers: list of regular expression patterns.
        :param dataframe: optional dataframe of already parsed arguments.
        """
        self.root_dir = Path(root_dir)
        self.parsers = parsers
        self.df = dataframe

    @classmethod
    def from_arguments(
        cls,
        root_dir: Path,
        arguments: List[Dict[str, str] | str] | Dict[str, str] | str,
        auto_sort: bool = True,
    ) -> ResultManager:
        """Create an instance of ResultManager from the given arguments.
        The single can be specified in two ways:
        - as a single string. In this case there will be no parsing and only folders
        with the specified name will be parsed
        - as a {key: value} dictionary. In this case, every key is the name of a
        parameter, and the value is a string specifying the regular expression for
        parsing.

        If a list of arguments is provided, this specifies a hierarchical folder
        structure, every level has parsing specified by the relative list argument.
        If auto_sort is set to True, within every dictionary the arguments are sorted in
        a canonical way.

        :param root_dir: root directory
        :param arguments: arguments, see `help(ResultManager)`
        :param auto_sort: whether to sort the arguments in the canonical order
        :return: a ArgumentManager instance
        """
        if not isinstance(arguments, list):
            arguments = [arguments]

        parsers = [
            re.compile(_pattern_from_arguments(arg, auto_sort)) for arg in arguments
        ]

        return ResultManager(root_dir, parsers)

    @classmethod
    def create_default_path(
        cls,
        root_dir: str | Path,
        values: List[Dict[str, Any] | str] | Dict[str, Any] | str,
        auto_sort: bool = True,
    ) -> Path:
        """Create the default path given

        :param root_dir: root directory
        :param values: the values specifying the directory structure.
            If it is a string, it specifies a simple directory name.
            If it is a dictionary, it specifies the {parameter name: value} that the
            directory name describes.
            If it is a list, it contains the string or dictionaries of values at every
            sub-level of root:dir.
        :param auto_sort: whether to sort the values specified as dictionaries according
            to the canonical order.
        :return: the path to a sub-folder with specified names.
        """
        root_dir = Path(root_dir)
        if not isinstance(values, list | tuple):
            values = [values]

        path = root_dir
        for arg in values:
            dir_name = (
                str(arg)
                if isinstance(arg, str | int | float)
                else _dirname_from_values(arg, auto_sort)
            )
            path = path / dir_name

        return path

    @property
    def patterns(self) -> List[str]:
        return [parser.pattern for parser in self.parsers]

    def parse_paths(self, **kwargs) -> None:
        """Recursively parse the root directory according to the specified parsers."""
        records = [
            {**{"__PATH__": res_path}, **match_group_args}
            for res_path, match_group_args in recursive_folder_parsing(
                self.root_dir, self.parsers
            )
        ]
        self.df = pd.DataFrame.from_records(records, **kwargs)

    def filter_results(
        self,
        equal: Optional[Dict[str, Any]] = None,
        contained: Optional[Dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """Filter the results in the result dataframe by row.
        The rows whose column values are equal and/or contained in those specified, are
        returned in the form of a new data frame. Notice that this method is different
        from pandas DataFrame's filter, which filters along both axes and only by
        column or row name.

        :param equal: dictionary of {column name: value} pairs.
            Rows with column value equal to that specified are returned.
        :param contained: dictionary of {column name: iterable of values} pairs:
            Rows with column values contained in those specified are returned.
        :return: a new data frame according to the specified filters.
        """
        filtered = self.df

        if equal is not None:
            for key, val in equal.items():
                mask = filtered[key] == val
                filtered = filtered[mask]

        if contained is not None:
            for key, val in contained.items():
                mask = filtered[key].isin(val)
                filtered = filtered[mask]

        return filtered


# Utility functions specifying how the ResultManager builds the patterns.
# We use the following name convention for the dictionaries of arguments and values:
# - arguments are utilized to build regular expression patterns. They consist of
#   {parameter name: string}, where the string are compiled to regular expression
#   patterns
# - values are concrete {parameter name: value} pairs.
def _pattern_from_arguments(arguments: Dict[str, str] | str, auto_sort=True) -> str:
    if isinstance(arguments, str):
        return arguments

    keys = _sorted_parameters(arguments) if auto_sort else arguments
    pattern = "_".join(_argument_pattern(key, arguments[key]) for key in keys)

    return pattern


def _dirname_from_values(
    values: Dict[str, str] | List[str] | str, auto_sort=True
) -> str:
    # if isinstance(values, list | tuple):
    #     return Path(*(_dirname_from_values(v) for v in values)).as_posix()
    if isinstance(values, str | int | float):
        return str(values)
    flat_dicts = _dict_to_multi_level(values)
    dirname = []
    for d in flat_dicts:
        keys = _sorted_parameters(d) if auto_sort else d
        dirname.append("_".join(_value_pattern(key, d[key]) for key in keys))
    return Path(*dirname).as_posix()


def _dict_to_multi_level(d: Dict):
    L = collections.defaultdict(dict)
    for kp, v in tree.iter_path(d):
        kstr = ".".join(kp)
        L[len(kp)][kstr] = v
    levels = sorted(L)
    return [L[i] for i in levels]


def _sorted_parameters(*params):
    return sorted(*params)


def _argument_pattern(argument: str, expr: str) -> str:
    pattern = f"{argument}=(?P<{argument}>{expr})"
    return pattern


def _value_pattern(argument: str, expr: Any) -> str:
    pattern = f"{argument}={expr}"
    return pattern
